#==================================================================================
# The code is ported from Bill Schottstedaet's 'clm-ins.scm' 
# file  available at https://ccrma.stanford.edu/software/snd/sndlib/
#==================================================================================


import math
import random
import cython
import numpy as np
import pysndlib.clm as clm
if cython.compiled:
    from cython.cimports.pysndlib import clm

# --------------- attract ---------------- #
#   ;; by James McCartney, from CMJ vol 21 no 3 p 6
@cython.ccall
def attract(beg, dur, amp, c):
    st: cython.long = clm.seconds2samples(beg)
    nd: cython.long = clm.seconds2samples(beg + dur)
    a: cython.double = .2
    b: cython.double = .2
    dt: cython.double = .04
    scale: cython.double = (.5 * amp) / c
    x: cython.double = -1.0
    y: cython.double = 0.0
    z: cython.double = 0.0
    x1: cython.double = 0.0
    
    i: cython.long = 0
    
    for i in range(st,nd):
        x1 = x - (dt * (y + z))
        y  += (dt * (x + (a * y)))
        z  += (dt * ((b + (x * z)) - (c * z)))
        x = x1
        clm.outa(i, scale * x)


if __name__ == '__main__': 
    with clm.Sound(play=True, statistics=True):
        attract(0, 2, .5, 4)



#==================================================================================
# The code is an attempt at translation of Bill Schottstedaet's 'env.scm' 
# file available at https://ccrma.stanford.edu/software/snd/sndlib/
#==================================================================================


from functools import singledispatch
import operator
import types
import math
import numpy as np
import numpy.typing as npt
cimport numpy as np
import cython
cimport cython
import pysndlib.clm as clm


@cython.ccall
def mapper(e1,e2,op):
    if isinstance(e1, list):
        return list(map(op, e1, e2))
    elif isinstance(e1, np.ndarray):
        return list(map(op, e1, e2))
    else:
        raise TypeError("expects list of ndarray")
    
@cython.ccall
def interleave(a,b):
    if isinstance(a, list):
        return [x for y in zip(a, b) for x in y]
    elif isinstance(a, np.ndarray):
        return np.vstack((a,b)).reshape((-1,),order='F')
    else:
        raise TypeError("expects list of ndarray")
    
    
@cython.ccall
def scale_and_offset(e, scl, offset: cython.double=0.0):
    if isinstance(e, list):
        return [x * scl + offset for x in e]
    elif isinstance(e, np.ndarray):
        return e * scl + offset

# TODO: --------------- window_envelopes ---------------- #    

# --------------- map_envelopes ---------------- #

@cython.ccall
def map_envelopes(func, e1, e2):
    """
    maps func over the breakpoints in env1 and env2 returning a new envelope
    map_envelopes(func, env1, env2) 
    """
    return mapper(e1,e2,func)

# --------------- multiply_envelopes ---------------- #
@cython.ccall
def multiply_envelopes(e1,e2):
    """
    multiplies break-points of env1 and env2 returning a new 
    envelope: multiply_envelopes([0,0,2,.5], [0,0,1,2,2,1]) -> [0,0,0.5,0.5,1.0,0.5]
    """
    return map_envelopes(operator.mul, e1, e2)
    
# --------------- add_envelopes ---------------- #
@cython.ccall
def add_envelopes(e1,e2):
    """
    adds break-points of env1 and env2 returning a new envelope"
    """

    return map_envelopes(operator.add, e1, e2)

# --------------- max-envelope ---------------- #
@cython.ccall
def max_envelope(e):
    """
    max_envelope(e) -> max y value in env
    """
    return max(e[1::2])
  
# --------------- min-envelope ---------------- #  
@cython.ccall
def min_envelope(e):
    """
    min_envelope(e) -> min y value in env
    """
    return min(e[1::2])
    
 
# --------------- integrate-envelope  ---------------- #    
# 
@cython.ccall
def integrate_envelope(e):
    """
    integrate_envelope(e) -> area under env
    """
    sum = 0.0
    for i in range(0,len(e)-2,2):
        sum += (e[i+1] + e[i+3]) * .5 * (e[i+2] - e[i])
    return sum

# --------------- envelope_last_x  ---------------- #  
@cython.ccall
def envelope_last_x(e):
    """
    envelope_last_x(env) -> max x axis break point position
    """
    return e[-2]


# ---------------  stretch_envelope   ---------------- #  
@cython.ccall
def stretch_envelope(fn, old_att, new_att, old_dec=None, new_dec=None):
    """
    takes 'env' and returns a new envelope based on it but with the attack and optionally decay portions stretched 
    or squeezed; 'old_att' is the original x axis attack end point, 'new_att' is where that 
    section should end in the new envelope.  Similarly for 'old_dec' and 'new_dec'.  This mimics 
    divseg in early versions of CLM and its antecedents in Sambox and Mus10 (linen).
    stretch_envelope([0,0,1,1], .1, .2) -> [0,0,0.2,0.1,1.0,1] 
    stretch_envelope([0,0,1,1,2,0], .1, .2, 1.5, 1.6) -> [0,0,0.2,0.1,1.1,1,1.6,0.5,2.0,0]
    """

    if not new_att and old_att:
        raise RuntimeError("new_att and old_att must be specified")
        
    else:
        if not new_att:
            return fn
        else:
            if old_dec and not new_dec:
                raise RuntimeError("old_dec and new_dec must be specified")
            
            else:
                x0 = fn[0]
                y0 = fn[1]
                new_x = x0
                last_x = fn[-2]
                new_fn = [x0,y0]
                scl = (new_att-x0) / max(.0001, old_att-x0)
                
                if old_dec and (old_dec == old_att):
                    old_dec =  .000001 * last_x
                
                for i in range(2,len(fn)-1, 2):
                    x1 = fn[i]
                    y1 = fn[i+1]
 
                    if x0 < old_att and x1 >= old_att:
                        y0 = y1 if x1==old_att else (y0 + ((y1 - y0) * ((old_att - x0) / (x1 - x0))))
                        
                        x0 = old_att

                        new_x = new_att
                        new_fn.extend((new_x,y0))   
                        scl = ((new_dec - new_att) / (old_dec - old_att)) if old_dec else ((last_x - new_att) / (last_x - old_att))
                        
                    if old_dec and x0 < old_dec and x1 >= old_dec:
                        y0 = y1 if x1 == old_dec else y0 + (y1 - y0) * ((old_dec - x0) / (x1 - x0))
                        x0 = old_dec
                        new_x = new_dec
                   
                        new_fn.extend((new_x,y0))
                        scl = (last_x - new_dec) / (last_x - old_dec)

                    if x0 != x1:
                        new_x = new_x + scl * (x1 - x0)
                        new_fn.extend((new_x,y1))
                        x0 = x1
                        y0 = y1
                        
                return new_fn
# --------------- scale_envelope  ---------------- # 
@cython.ccall
def scale_envelope(e, scaler: cython.double, offset: cython.double=0.0):
    """
    scales y axis values by 'scaler' and optionally adds 'offset'
    """
    ys = scale_and_offset(e[1::2], scaler, offset)
    xs = e[0::2]
    return interleave(xs,ys)
    
# --------------- reverse_envelope  ---------------- # 
@cython.ccall
def reverse_envelope(e):
    """
    reverses the breakpoints in env
    """
    xs = e[0::2]
    ys = e[:0:-2]
    return interleave(xs, ys)

# --------------- concatenate_envelopes  ---------------- # 


def concatenate_envelopes(*args, epsilon: cython.double =.01):
    """
    concatenates its arguments into a new envelope
    """
    # get list of last_x of each env
    # also verify each envelope
    res = None
    
    for e in args:
        clm.validate_envelope(e)
        
        if res is None:
            res = np.array(e)
        else:    
            tmp = np.array(e)
            tmp[0::2] += res[-2]
            if tmp[0] == res[-2]:
                tmp[0] += epsilon
            res = np.append(res, tmp)
            
    return list(res)  
    
# --------------- normalize_envelope  ---------------- # 
@cython.ccall
def normalize_envelope(e, new_max: cython.double=1.0):
    maxy = max_envelope(e)
    scl = new_max / maxy
    return scale_envelope(e, scl)
    
    



# cython: c_string_type=unicode
# cython: c_string_encoding=utf8

import numpy as np
cimport cython
cimport numpy as np
cimport pysndlib.csndlib as csndlib

csndlib.mus_sound_initialize()

cpdef enum Header:
    """
    file types for audio files
    """
    UNKNOWN_HEADER, NEXT, AIFC, RIFF, RF64, BICSF, NIST, INRS, ESPS, SVX, VOC, SNDT, RAW, SMP, AVR, IRCAM, SD1, SPPACK, MUS10, HCOM, PSION, MAUD, IEEE, MATLAB, ADC, MIDI, SOUNDFONT, GRAVIS, COMDISCO, GOLDWAVE, SRFS, MIDI_SAMPLE_DUMP, DIAMONDWARE, ADF, SBSTUDIOII, DELUSION, FARANDOLE, SAMPLE_DUMP, ULTRATRACKER, YAMAHA_SY85, YAMAHA_TX16W, DIGIPLAYER, COVOX, AVI, OMF, QUICKTIME, ASF, YAMAHA_SY99, KURZWEIL_2000, AIFF, PAF, CSL, FILE_SAMP, PVF, SOUNDFORGE, TWINVQ, AKAI4, IMPULSETRACKER, KORG, NVF, CAFF, MAUI, SDIF, OGG, FLAC, SPEEX, MPEG, SHORTEN, TTA, WAVPACK, SOX, NUM_HEADERS,

cpdef enum Sample:
    """
    numerical sample types
    """
    UNKNOWN_SAMPLE, BSHORT, MULAW, BYTE, BFLOAT, BINT, ALAW, UBYTE, B24INT, BDOUBLE, LSHORT, LINT, LFLOAT, LDOUBLE, UBSHORT, ULSHORT, L24INT, BINTN, LINTN, BFLOAT_UNSCALED, LFLOAT_UNSCALED, BDOUBLE_UNSCALED, LDOUBLE_UNSCALED, NUM_SAMPLES,

cpdef enum Error:
    """
    sndlib and clm errors
    """
    NO_ERROR, NO_FREQUENCY, NO_PHASE, NO_GEN, NO_LENGTH, NO_DESCRIBE, NO_DATA, NO_SCALER, MEMORY_ALLOCATION_FAILED, CANT_OPEN_FILE, NO_SAMPLE_INPUT, NO_SAMPLE_OUTPUT, NO_SUCH_CHANNEL, NO_FILE_NAME_PROVIDED, NO_LOCATION, NO_CHANNEL, NO_SUCH_FFT_WINDOW, UNSUPPORTED_SAMPLE_TYPE, HEADER_READ_FAILED, UNSUPPORTED_HEADER_TYPE, FILE_DESCRIPTORS_NOT_INITIALIZED, NOT_A_SOUND_FILE, FILE_CLOSED, WRITE_ERROR, HEADER_WRITE_FAILED, CANT_OPEN_TEMP_FILE, INTERRUPTED, BAD_ENVELOPE, AUDIO_CHANNELS_NOT_AVAILABLE, AUDIO_SRATE_NOT_AVAILABLE, AUDIO_SAMPLE_TYPE_NOT_AVAILABLE, AUDIO_NO_INPUT_AVAILABLE, AUDIO_CONFIGURATION_NOT_AVAILABLE, AUDIO_WRITE_ERROR, AUDIO_SIZE_NOT_AVAILABLE, AUDIO_DEVICE_NOT_AVAILABLE, AUDIO_CANT_CLOSE, AUDIO_CANT_OPEN, AUDIO_READ_ERROR, AUDIO_CANT_WRITE, AUDIO_CANT_READ, AUDIO_NO_READ_PERMISSION, CANT_CLOSE_FILE, ARG_OUT_OF_RANGE, NO_CHANNELS, NO_HOP, NO_WIDTH, NO_FILE_NAME, NO_RAMP, NO_RUN, NO_INCREMENT, NO_OFFSET, NO_XCOEFF, NO_YCOEFF, NO_XCOEFFS, NO_YCOEFFS, NO_RESET, BAD_SIZE, CANT_CONVERT, READ_ERROR, NO_FEEDFORWARD, NO_FEEDBACK, NO_INTERP_TYPE, NO_POSITION, NO_ORDER, NO_COPY, CANT_TRANSLATE, NUM_ERRORS,



cpdef cython.long mus_sound_samples(str file)
cpdef cython.long mus_sound_framples(str file)
cpdef cython.int mus_sound_datum_size(str file)
cpdef cython.long mus_sound_data_location(str file)
cpdef cython.int mus_sound_chans(str file)
cpdef cython.double mus_sound_srate(str file)
cpdef Header mus_sound_header_type(str file)
cpdef Sample mus_sound_sample_type(str file)
cpdef Sample mus_sound_original_sample_type(str file)
cpdef cython.long mus_sound_comment_start(str file)

cpdef cython.long mus_sound_comment_end(str file)
cpdef cython.long mus_sound_length(str file)
cpdef cython.long mus_sound_fact_samples(str file)
cpdef cython.int mus_sound_type_specifier(str file)
cpdef cython.int mus_sound_block_align(str file)
cpdef cython.int mus_sound_bits_per_sample(str file)
cpdef cython.int mus_sound_set_chans(str file,  cython.int val)
cpdef cython.int mus_sound_set_srate(str file, cython.int val)
cpdef Header mus_sound_set_header_type(str file,  csndlib.mus_header_t mus_header)
cpdef Sample mus_sound_set_sample_type(str file, csndlib.mus_sample_t mus_sample)
cpdef cython.int mus_sound_set_data_location(str file, csndlib.mus_long_t val)
cpdef cython.int mus_sound_set_samples(str file, csndlib.mus_long_t val)
cpdef str mus_header_type_name(csndlib.mus_header_t header_type)
cpdef str mus_header_type_to_string(Header header_type)
cpdef str  mus_sample_type_name( csndlib.mus_sample_t samp_type)
cpdef str mus_sample_type_to_string(Sample samp_type)
cpdef str mus_sample_type_short_name(Sample samp_type)
cpdef str mus_sound_comment(str file)
cpdef cython.int mus_bytes_per_sample(Sample samp_type)
cpdef cython.double mus_sound_duration(str file)
cpdef cython.int mus_sound_initialize()
cpdef cython.int mus_sound_override_header(str file, cython.int srate, cython.int chans, Sample sample_type, Header header_type, cython.int location, cython.int size: int )
cpdef np.ndarray mus_sound_loop_info(str filename)
cpdef mus_sound_set_loop_info(str filename, np.ndarray info)
cpdef cython.int  mus_sound_open_input(str filename)
cpdef cython.int  mus_sound_open_output(str file, cython.int srate, cython.int chans, Sample sample_type, Header header_type, str comments)
cpdef cython.int  mus_sound_reopen_output(str file, cython.int srate, cython.int chans, Sample sample_type, Header header_type,  cython.int  location)
cpdef cython.int mus_sound_close_input(cython.int fd)
cpdef cython.int mus_sound_close_output(cython.int fd, csndlib.mus_long_t bytes_of_data)







import math
import random
#==================================================================================
# The code is ported from Bill Schottstedaet's 'clm-ins.scm' 
# file  available at https://ccrma.stanford.edu/software/snd/sndlib/
#==================================================================================

import cython
import numpy as np
import pysndlib.clm as clm
import pysndlib.env as env
if cython.compiled:
    from cython.cimports.pysndlib import clm


# --------------- fm_drum ---------------- #
# ;;; Jan Mattox's fm drum:

@cython.ccall
def fm_drum(start_time, duration, frequency, amplitude, index, high=True, degree=0.0, distance=1.0, reverb_amount=.01):
    indxfun = [0,0,5,.014,10,.033,15,.061,20,.099,
            	25,.153,30,.228,35,.332,40,.477,
		        45,.681,50,.964,55,.681,60,.478,65,.332,
		        70,.228,75,.153,80,.099,85,.061,
		        90,.033,95,.0141,100,0]

    indxpt: cython.double = 100 - (100 * ((duration - .1) / duration))
    atdrpt: cython.double = 100 * ((.01 if high else .015) / duration)
    divindxf = env.stretch_envelope(indxfun, 50, atdrpt, 65, indxpt)
    ampfun = [0,0,3,.05,5,.2,7,.8,8,.95,10,1.0,12,.95,20,.3,30,.1,100,0]
    casrat: cython.double = 8.525 if high else 3.515    
    fmrat: cython.double = 3.414 if high else 1.414
    glsfun = [0,0,25,0,75,1,100,1]
    beg: cython.long = clm.seconds2samples(start_time)
    end: cython.long = clm.seconds2samples(start_time + duration)
    glsf = clm.make_env(glsfun, scaler=(clm.hz2radians(66) if high else 0.0), duration=duration)
    ampe = env.stretch_envelope(ampfun, 10, atdrpt, 15, max((atdrpt + 1), 100 - (100 * ((duration - .05) / duration))))
    ampf  = clm.make_env(ampe, scaler=amplitude, duration=duration)
    indxf  = clm.make_env(divindxf, scaler=min(clm.hz2radians(index * fmrat * frequency), math.pi), duration=duration)
    mindxf  = clm.make_env(divindxf, scaler=min(clm.hz2radians(index * casrat * frequency), math.pi), duration=duration)
    deve = env.stretch_envelope(ampfun, 10, atdrpt, 90, max((atdrpt + 1), (100 - (100 * ((duration - .05) / duration)))))
    devf  = clm.make_env(deve, scaler=min(math.pi, clm.hz2radians(7000)), duration=duration)
    loc  = clm.make_locsig(degree, distance, reverb_amount)
    rn  = clm.make_rand(7000., 1.0)
    carrier  = clm.make_oscil(frequency)
    fmosc  = clm.make_oscil(frequency*fmrat)
    cascade = clm.make_oscil(frequency*casrat)
    
    i: cython.long = 0
    gls: cython.double = 0.0
    
    for i in range(beg, end):
        gls = clm.env(glsf)
        clm.locsig(loc, i, (clm.env(ampf) * clm.oscil(carrier, (gls + (clm.env(indxf) * 
                                        clm.oscil(fmosc, ((gls*fmrat)+(clm.env(mindxf)*
                                            clm.oscil(cascade, ((gls*casrat)+(clm.env(devf)*clm.rand(rn))))))))))))
if __name__ == '__main__':  
    with clm.Sound(play=True, statistics=True):
        fm_drum(0, 1.5, 55, .3, 5, False)
        fm_drum(2, 1.5, 66, .3, 5, True)



#==================================================================================
# The code is ported from Bill Schottstedaet's 'clm-ins.scm' 
# file  available at https://ccrma.stanford.edu/software/snd/sndlib/
#==================================================================================

import math
import random
import cython
import numpy as np
import pysndlib.clm as clm
import pysndlib.env as env
if cython.compiled:
    from cython.cimports.pysndlib import clm

# --------------- cellon ---------------- #

@cython.ccall
def cellon(beg, dur, pitch0, amp, ampfun, betafun, beta0, beta1, betaat, betadc, ampat, ampdc, dis, pcrev, deg,
            pitch1, glissfun, glissat, glissdc, pvibfreq, pvibpc, pvibfun=[0,1,100,1], pvibat=0, pvibdc=0, rvibfreq=0, rvibpc=0, rvibfun=[0,1,100,1]):
	
    st: cython.long = clm.seconds2samples(beg)
    nd: cython.long = clm.seconds2samples(beg + dur)
    pit1 = pitch0 if pitch1 == 0.0 else pitch1
    loc = clm.make_locsig(deg, dis, pcrev)
    carrier = clm.make_oscil(pitch0)
    low = clm.make_one_zero(.5, -.5)
    fm = 0.0
    fmosc = clm.make_oscil(pitch0)
    pvib = clm.make_triangle_wave(pvibfreq, 1.0)
    rvib = clm.make_rand_interp(rvibfreq, 1.0)
    ampap: cython.double = (100 * (ampat / dur)) if ampat > 0.0 else 25
    ampdp: cython.double = (100 * (1.0 - (ampdc / dur))) if ampdc > 0.0 else 75
    glsap: cython.double = (100 * (glissat / dur)) if glissat > 0.0 else 25
    glsdp: cython.double = (100 * (1.0 - (glissdc / dur))) if glissdc > 0.0 else 75
    betap: cython.double = (100 * (betaat / dur)) if betaat > 0.0 else 25
    betdp: cython.double = (100 * (1.0 - (betadc / dur))) if betadc > 0.0 else 75
    pvbap: cython.double = (100 * (pvibat / dur)) if pvibat > 0.0 else 25
    pvbdp: cython.double = (100 * (1.0 - (pvibdc / dur))) if pvibdc > 0.0 else 75
    
    pvibenv = clm.make_env(env.stretch_envelope(pvibfun, 25, pvbap, 75, pvbdp), duration=dur, scaler=clm.hz2radians(pvibpc * pitch0))
    rvibenv = clm.make_env(rvibfun, duration=dur, scaler=clm.hz2radians(rvibpc * pitch0))
    glisenv = clm.make_env(env.stretch_envelope(pvibfun, 25, pvbap, 75, pvbdp), duration=dur, scaler=clm.hz2radians(pvibpc * pitch0))
    amplenv = clm.make_env(env.stretch_envelope(ampfun, 25, ampap, 75, ampdp), scaler=amp, duration=dur)
    betaenv = clm.make_env(env.stretch_envelope(betafun, 25, betap, 75, betdp), duration=dur, scaler=(beta1 - beta0), offset=beta0)
    
    i: cython.long = 0
    fm: cython.double = 0.0
    vib: cython.double = 0.0
    if (pitch0 == pitch1) and (clm.is_zero(pvibfreq) or clm.is_zero(pvibpc)) and (clm.is_zero(rvibfreq) or clm.is_zero(rvibpc)):
        for i in range(st, nd):
            fm = clm.one_zero(low, clm.env(betaenv) * clm.oscil(fmosc, fm))
            clm.locsig(loc, i, clm.env(amplenv)*clm.oscil(carrier,fm))
    else:
        for i in range(st, nd):
            vib = (clm.env(pvibenv)*clm.triangle_wave(pvib)) + (clm.env(rvibenv)*clm.rand_interp(rvib)) + clm.env(glisenv)
            fm = clm.one_zero(low, clm.env(betaenv) * clm.oscil(fmosc, (fm+vib)))
            clm.locsig(loc, i, clm.env(amplenv)*clm.oscil(carrier,fm+vib))


if __name__ == '__main__': 
    with clm.Sound( play = True, statistics=True):
        cellon(0,1,220,.1,[0,0,25,1,75,1,100,0],[0,0,25,1,75,1,100,0],.75,1.0,0,0,0,0,1,0,0,220,[0,0,25,1,75,1,100,0],0,0,0,0,[0,0,100,0],0,0,0,0,[0,0,100,0])  



#==================================================================================
# The code is ported from Bill Schottstedaet's 'jcrev.scm' 
# file  available at https://ccrma.stanford.edu/software/snd/sndlib/
#==================================================================================
import math
import cython
import numpy as np
import pysndlib.clm as clm
import pysndlib.env as env
if cython.compiled:
    from cython.cimports.pysndlib import clm

@clm.clm_reverb
def jc_reverb(lowpass=False, volume=1., amp_env = None, decay_time=1.0):
   
    allpass1 = clm.make_all_pass(-.7, .7, 1051)
    allpass2 = clm.make_all_pass(-.7, .7, 337)
    allpass3 = clm.make_all_pass(-.7, .7, 113)
    comb1 = clm.make_comb(.742, 4799)
    comb2 = clm.make_comb(.733, 4999)
    comb3 = clm.make_comb(.715, 5399)
    comb4 = clm.make_comb(.697, 5801)
    chans = clm.get_channels(clm.default.output)
    
    length: cython.long = math.floor(clm.get_length(clm.default.reverb) + (clm.get_srate()*decay_time)) # in for loop
    filts = [clm.make_delay(clm.seconds2samples(.013))] if chans == 1 else [clm.make_delay(clm.seconds2samples(.013)),clm.make_delay(clm.seconds2samples(.011)) ]
    combs = clm.make_comb_bank([comb1, comb2, comb3, comb4])
    allpasses = clm.make_all_pass_bank([allpass1,allpass2,allpass3])
    
    i: cython.long = 0 # in for loop
    
    if lowpass or amp_env:
        flt = clm.make_fir_filter(3, [.25, .5, .25]) if lowpass else None
        envA = clm.make_env(amp_env, scaler=volume, duration = length / clm.get_srate())
        
        if lowpass:
            for i in range(length):
                clm.out_bank(filts, i, (clm.env(envA) * clm.fir_filter(flt, clm.comb_bank(combs, clm.all_pass_bank(allpasses, clm.ina(i, clm.default.reverb))))))
        else:
            for i in range(length):
                clm.out_bank(filts, i, (clm.env(envA) * clm.comb_bank(combs, clm.all_pass_bank(allpasses, clm.ina(i, clm.default.reverb)))))
    else:
        if chans == 1:
            
            gen = filts[0]
            for i in range(length):
                clm.outa(i, clm.delay(gen, volume * clm.comb_bank(combs, clm.all_pass_bank(allpasses, clm.ina(i, clm.default.reverb)))))
        else:    
            gen1: clm.mus_any = filts[0]
            gen2: clm.mus_any = filts[1]
            for i in range(length):
                val: cython.double = volume * clm.comb_bank(combs, clm.all_pass_bank(allpasses, clm.ina(i, clm.default.reverb))) 
                clm.outa(i, clm.delay(gen1, val))
                clm.outb(i, clm.delay(gen2, val))



#==================================================================================
# The code is ported from Bill Schottstedaet's 'clm-ins.scm' 
# file available at https://ccrma.stanford.edu/software/snd/sndlib/
#==================================================================================

import math
import random
import cython
import numpy as np
import pysndlib.clm as clm
import pysndlib.env as env
if cython.compiled:
    from cython.cimports.pysndlib import clm

# --------------- gong ---------------- #
# ;;; Paul Weineke's gong.
@cython.ccall
def gong(start_time, duration, frequency, amplitude, degree=0.0, distance=1.0, reverb_amount=.005):
    mfq1 = frequency * 1.16
    mfq2 = frequency * 3.14
    mfq3 = frequency * 1.005    
    indx01 = clm.hz2radians(.01 * mfq1)
    indx11 = clm.hz2radians(.30 * mfq1)
    indx02 = clm.hz2radians(.01 * mfq2)
    indx12 = clm.hz2radians(.38 * mfq2)
    indx03 = clm.hz2radians(.01 * mfq3)
    indx13 = clm.hz2radians(.50 * mfq3)
    atpt = 5
    atdur = 100 * (.002 / duration)
    expf = [0,0,3,1,15,.5,27,.25,50,.1,100,0]
    rise = [0,0,15,.3,30,1.0,75,.5,100,0]
    fmup = [0,0,75,1.0,98,1.0,100,0]
    fmdwn = [0,0,2,1.0,100,0]
    ampfun = clm.make_env(env.stretch_envelope(expf, atpt, atdur, None, None), scaler=amplitude, duration=duration)
    indxfun1 = clm.make_env(fmup, duration=duration, scaler=indx11-indx01, offset=indx01)
    indxfun2 = clm.make_env(fmup, duration=duration, scaler=indx12-indx02, offset=indx02)
    indxfun3 = clm.make_env(fmup, duration=duration, scaler=indx13-indx03, offset=indx03)
    loc = clm.make_locsig(degree, distance, reverb_amount)
    carrier = clm.make_oscil(frequency)
    mod1 = clm.make_oscil(mfq1)
    mod2 = clm.make_oscil(mfq2)
    mod3 = clm.make_oscil(mfq3)
    beg: cython.long = clm.seconds2samples(start_time)
    end: cython.long = clm.seconds2samples(start_time + duration)
    
    i: cython.long = 0
    
    for i in range(beg, end):
        clm.locsig(loc, i, clm.env(ampfun) * clm.oscil(carrier, (clm.env(indxfun1) * clm.oscil(mod1)) + (clm.env(indxfun2) * clm.oscil(mod2)) + (clm.env(indxfun3) * clm.oscil(mod3))))
    
if __name__ == '__main__':     
    with clm.Sound(play=True, statistics=True):
        gong(0, 3, 261.61, .6)
    



#==================================================================================
# The code is ported from Bill Schottstedaet's 'clm-ins.scm' 
# file available at https://ccrma.stanford.edu/software/snd/sndlib/
#==================================================================================

import math
import random
import cython
import numpy as np
import pysndlib.clm as clm
if cython.compiled:
    from cython.cimports.pysndlib import clm

# --------------- scratch ---------------- #

# Made changes based on callback functions and such in python


def scratch(start, file, src_ratio, turnaroundlist):
    f = clm.make_file2frample(file)
    beg = clm.seconds2samples(start)
    turntable = turnaroundlist
    turn_i = 1
    turns = len(turnaroundlist)
    cur_sample = clm.seconds2samples(turntable[0])
    
    turn_sample = clm.seconds2samples(turntable[1])
    
    def func(direction):
        nonlocal cur_sample
        inval = clm.file2sample(f, cur_sample)
        cur_sample = cur_sample + direction
        return inval
   
    turning = 0
    last_val = 0.0
    last_val2 = 0.0
    rd = clm.make_src(func, srate=src_ratio)
    forwards = src_ratio > 0.0
    
    if forwards and turn_sample < cur_sample:
        rd.mus_increment = -src_ratio
    
    i: cython.long = beg
    
    while turn_i < turns:

        val = clm.src(rd, 0.0)

        if turning == 0:
            if forwards and (cur_sample >= turn_sample):
                turning = 1
            else:
                if (not forwards) and (cur_sample <= turn_sample):
                    turning = -1
        else:                
            if ((last_val2 <= last_val) and (last_val >= val)) or ((last_val2 >= last_val) and (last_val <= val)):
                turn_i += 1
                if turn_i < turns:
                    turn_sample = clm.seconds2samples(turntable[turn_i])
                    forwards = not forwards
                    rd.mus_increment = -rd.mus_increment
                turning = 0
        last_val2 = last_val
        last_val = val
        clm.outa(i, val)
        i += 1

if __name__ == '__main__': 
    with clm.Sound( play = True, statistics=True):
        scratch(0.0, '../examples/yeah.aiff', 1., [0.0, .5, .25, 1.0, .5, 5.0])    


                   
                    



#==================================================================================
# The code is ported from Bill Schottstedaet's 'clm-ins.scm' 
# file  available at https://ccrma.stanford.edu/software/snd/sndlib/
#==================================================================================

import math
import random
import cython
import numpy as np
import pysndlib.clm as clm
if cython.compiled:
    from cython.cimports.pysndlib import clm



# --------------- fm_bell ---------------- #
@cython.ccall
def fm_bell(startime, dur, frequency, amplitude, 
        amp_env=[0, 0, .1, 1, 10, .6, 25, .3, 50, .15, 90, .1, 100, 0], 
        index_env=[0, 1, 2, 1.1, 25, .75, 75, .5, 100, .2], index=1.):
    fmInd2: cython.double = clm.hz2radians( (8.0 - (frequency / 50.)) * 4)
    beg: cython.long = clm.seconds2samples(startime)
    end: cython.long = clm.seconds2samples(startime + dur)
    fmInd1: cython.double = clm.hz2radians(32.0 * frequency)
    fmInd3: cython.double = fmInd2 * .705 * (1.4 - (frequency / 250.))
    fmInd4: cython.double = clm.hz2radians(32.0 * (20  - (frequency / 20)))
    mod1 = clm.make_oscil(frequency * 2)
    mod2 = clm.make_oscil(frequency * 1.41)
    mod3 = clm.make_oscil(frequency * 2.82)
    mod4 = clm.make_oscil(frequency * 2.4)
    car1 = clm.make_oscil(frequency)
    car2 = clm.make_oscil(frequency)
    car3 = clm.make_oscil(frequency * 2.4)
    indf = clm.make_env(index_env, scaler=index, duration=dur)
    ampf = clm.make_env(amp_env, scaler=amplitude, duration=dur)
    
    i: cython.long = 0
    fmenv: cython.double = 0.0
    
    for i in range(beg, end):
        fmenv = clm.env(indf)
        clm.outa(i, clm.env(ampf) * (clm.oscil(car1, fm=(fmenv * fmInd1 * clm.oscil(mod1))) +
                             (.15 * clm.oscil(car2, fm=(fmenv * ( (fmInd2 * clm.oscil(mod2)) + (fmInd3 * clm.oscil(mod3)) )))) +
                             (.15 * clm.oscil(car3, fm=(fmenv * fmInd4 * clm.oscil(mod4)))) ))
  
if __name__ == '__main__':                             
    with clm.Sound( play = True, statistics=True, channels=1 ):
        fbell = [0,1,2,1.1,25,.75,75,.5,100,.2]
        abell = [0,0,.1,1,10,.6,25,.3,50,.15,90,.1,100,0]
        fm_bell( 0,5.000,233.046,.028,abell,fbell,.750)
        fm_bell( 5.912,2.000,205.641,.019,abell,fbell,.650)
        fm_bell( 6.085,5.000,207.152,.017,abell,fbell,.750)
        fm_bell( 6.785,7.000,205.641,.010,abell,fbell,.650)
        fm_bell( 15.000,.500,880,.060,abell,fbell,.500)
        fm_bell( 15.006,6.500,293.66,.1,abell,fbell,0.500)
        fm_bell( 15.007,7.000,146.5,.1,abell,fbell,1.000)
        fm_bell( 15.008,6.000,110,.1,abell,fbell,1.000)
        fm_bell( 15.010,10.00,73.415,.028,abell,fbell,0.500)
        fm_bell( 15.014,4.000,698.46,.068,abell,fbell,0.500)
        



#==================================================================================
# The code is ported from Bill Schottstedaet's 'clm-ins.scm' 
# file  available at https://ccrma.stanford.edu/software/snd/sndlib/
#==================================================================================

import math
import random
import cython
import numpy as np
import pysndlib.clm as clm
import pysndlib.env as env
if cython.compiled:
    from cython.cimports.pysndlib import clm

# --------------- pqw ---------------- #
#   ;; phase-quadrature waveshaping used to create asymmetric (i.e. single side-band) spectra.
#   ;; The basic idea here is a variant of sin x sin y - cos x cos y = cos (x + y)

def clip_env(e):
    xs = e[0::2]
    ys = [min(x,1.0) for x in e[1::2]]
    return env.interleave(xs, ys)
    
    
@cython.ccall
def pqw(start, dur, spacing_freq, carrier_freq, amplitude, 
        ampfun, indexfun, partials, degree=0.0, distance=1.0, reverb_amount=.005):
    normalized_partials = clm.normalize_partials(partials)
    spacing_cos = clm.make_oscil(spacing_freq, initial_phase=math.pi/2.0)
    spacing_sin = clm.make_oscil(spacing_freq)
    carrier_cos = clm.make_oscil(carrier_freq, initial_phase=math.pi/2.0)
    carrier_sin = clm.make_oscil(carrier_freq)
    sin_coeffs: np.ndarray = clm.partials2polynomial(normalized_partials, clm.Polynomial.SECOND_KIND)
    cos_coeffs: np.ndarray = clm.partials2polynomial(normalized_partials, clm.Polynomial.FIRST_KIND)
    amp_env = clm.make_env(ampfun, scaler=amplitude, duration=dur)
    ind_env = clm.make_env(clip_env(indexfun), duration=dur)
    loc = clm.make_locsig(degree, distance, reverb_amount)
    r: cython.double = carrier_freq / spacing_freq
    tr = clm.make_triangle_wave(5, clm.hz2radians(.005 * spacing_freq))
    rn = clm.make_rand_interp(12, clm.hz2radians(.005 * spacing_freq))
    beg: cython.long = clm.seconds2samples(start)
    end: cython.long = clm.seconds2samples(start + dur)
    
    i: cython.long = 0
    vib: cython.double = 0.0
    
    for i in range(beg, end):
        vib = clm.triangle_wave(tr) + clm.rand_interp(rn)
        ax = clm.env(ind_env) * clm.oscil(spacing_cos, vib)
        clm.locsig(loc, i, clm.env(amp_env) * 
            ((clm.oscil(carrier_sin, (vib * r)) * clm.oscil(spacing_sin, vib) * clm.polynomial(sin_coeffs, ax)) -
                (clm.oscil(carrier_cos, (vib * r)) * clm.polynomial(cos_coeffs, ax))))

if __name__ == '__main__': 
    with clm.Sound( play = True, statistics=True):
        pqw(0, .5, 200, 1000, .2, [0,0,25,1,100,0], [0, 1, 100, 0], [2, .1, 3, .3, 6, .5])



#==================================================================================
# The code is ported from Bill Schottstedaet's 'clm-ins.scm' 
# file  available at https://ccrma.stanford.edu/software/snd/sndlib/
#==================================================================================

import math
import random
import cython
import numpy as np
import pysndlib.clm as clm
if cython.compiled:
    from cython.cimports.pysndlib import clm


# --------------- jl_reverb ---------------- #
@cython.ccall
@clm.reverb
def jl_reverb(decay_time: cython.double  =3.0, volume: cython.double =1.0):
    allpass1 = clm.make_all_pass(-.700, .700, 2111)
    allpass2 = clm.make_all_pass(-.700, .700, 673)
    allpass3 = clm.make_all_pass(-.700, .700, 223)
    comb1 = clm.make_comb(.742, 9601)
    comb2 = clm.make_comb(.733, 10007)
    comb3 = clm.make_comb(.715, 10799)
    comb4 = clm.make_comb(.697, 11597)
    outdel1 = clm.make_delay(clm.seconds2samples(.013))
    outdel2 = clm.make_delay(clm.seconds2samples(.011))
    length = math.floor((decay_time * clm.get_srate()) + clm.get_length(clm.default.reverb))
    filts = [outdel1, outdel2]
    combs = clm.make_comb_bank([comb1, comb2, comb3, comb4])
    allpasses = clm.make_all_pass_bank([allpass1, allpass2, allpass3])
    
    i: cython.long = 0
    
    for i in range(length):
        clm.out_bank(filts, i, volume * clm.comb_bank(combs, clm.all_pass_bank(allpasses, clm.ina(i, clm.default.reverb))))
        



import math
import random
import cython
import numpy as np
import pysndlib.clm as clm
import pysndlib.generators as gens
if cython.compiled:
    from cython.cimports.pysndlib import clm
    from cython.cimports.pysndlib import generators as gens
# from generators.scm 
# uses nkssb generator


# --------------- nkssber ---------------- #
@cython.ccall
def nkssber(beg, dur, freq, mfreq, n, vibfreq, amp):
    start: cython.long = clm.seconds2samples(beg)
    stop: cython.long = clm.seconds2samples(beg + dur)
    gen = gens.make_nkssb(freq, (mfreq / freq), n)
    move = clm.make_env([0,1,1,-1], duration=dur)
    vib = clm.make_polywave(vibfreq, [1,clm.hz2radians((freq / mfreq) * 5.)], clm.Polynomial.SECOND_KIND)
    ampf = clm.make_env([0,0,1,1,5,1,6,0], scaler=amp, duration=dur)
    i: cython.long = 0
    for i in range(start, stop):
        clm.outa(i, clm.env(ampf) * gens.nkssb_interp(gen, clm.polywave(vib), clm.env(move)))

if __name__ == '__main__':  
    with clm.Sound(play=True, statistics=True, scaled_to=.5):
        nkssber(0,1,1000,100,5,5,.5)
        nkssber(1,2,600,100,4,1,.5)
        nkssber(3,2,1000,540,3,3,.5)
        nkssber(5,4,300,120,2,.25,.5)
        nkssber(9,1,30,4,40,.5,.5)
        nkssber(10,1,20,6,80,.5,.5)



#==================================================================================
# The code is an attempt at translation of Bill Schottstedaet's 'dsp.scm' 
# file available at https://ccrma.stanford.edu/software/snd/sndlib/
#==================================================================================

# TODO: 'easy case '  

import math
import cython
import pysndlib.clm as clm
import numpy as np
if cython.compiled:
    from cython.cimports.pysndlib import clm


#not sure typing any of the args are worth it.
@cython.ccall
def fm_violin(beg, dur, frequency, amplitude, fm_index=1, 
            amp_env = [ 0, 0,  25, 1,  75, 1,  100, 0],
            periodic_vibrato_rate = 5.0,
            random_vibrato_rate = 16.0,
            periodic_vibrato_amplitude = 0.0025,
            random_vibrato_amplitude = 0.005,
            fm1_index=None,
            fm2_index=None,
            fm3_index=None,
            fm1_rat: cython.double = 1.0, # in for loop
            fm2_rat: cython.double = 3.0, # in for loop
            fm3_rat: cython.double = 4.0, # in for loop
            fm1_env = [0., 1., 25., .4, 75., .6, 100., 0.],
            fm2_env = [0., 1., 25., .4, 75., .6, 100., 0.],
            fm3_env = [0., 1., 25., .4, 75., .6, 100., 0.],
            gliss_env = [0,0,100,0],
            glissando_amount = 0.0,
            noise_amount = 0.0,
            noise_freq = 1000.,
            ind_noise_freq = 10.,
            ind_noise_amount = 0.0,
            amp_noise_freq = 20.0,
            amp_noise_amount = 0.0,
            degree=45,
            distance=1.0,
            reverb_amount = .01,
            base=1):   
    start: cython.long = clm.seconds2samples(beg) # in for loop
    end: cython.long = start + clm.seconds2samples(dur) # in for loop
    
    frq_scl = clm.hz2radians(frequency)
    maxdev = frq_scl * fm_index
    index1 = fm1_index if fm1_index else min(math.pi, maxdev * (5.0 / math.log(frequency)))
    index2 = fm2_index if fm2_index else min(math.pi, maxdev * 3.0 * ((8.5 - math.log(frequency)) / (3.0 + (frequency / 1000.))))
    index3 = fm3_index if fm3_index else min(math.pi, maxdev * (4.0 / math.sqrt(frequency)))
    carrier = clm.make_oscil(frequency)
    fmosc1 = clm.make_oscil(frequency)
    fmosc2 = clm.make_oscil(frequency * fm2_rat)
    fmosc3 = clm.make_oscil(frequency * fm3_rat)
    ampf = clm.make_env(amp_env, scaler = amplitude, duration = dur, base=base)
    indf1 = clm.make_env(fm1_env, scaler = index1, duration = dur)
    indf2 = clm.make_env(fm2_env, scaler = index2, duration = dur)
    indf3 = clm.make_env(fm3_env, scaler = index3, duration = dur)
    frqf = clm.make_env(gliss_env,scaler = (glissando_amount * frq_scl), duration=dur)
    pervib = clm.make_triangle_wave(periodic_vibrato_rate,periodic_vibrato_amplitude * frq_scl)
    ranvib = clm.make_rand_interp(random_vibrato_rate, random_vibrato_amplitude * frq_scl)
    fm_noi = clm.make_rand(noise_freq, math.pi * noise_amount)
    ind_noi = clm.make_rand_interp(ind_noise_freq, ind_noise_amount)
    amp_noi = clm.make_rand_interp(amp_noise_freq, amp_noise_amount)
    loc = clm.make_locsig(degree, distance, reverb_amount)
    
    i: cython.int = 0 # in for loop
    
    for i in range(start, end):
        vib: cython.double = clm.triangle_wave(pervib) + clm.rand_interp(ranvib) + clm.env(frqf)
        fuzz: cython.double = clm.rand(fm_noi)
        inoi: cython.double = 1.0 + clm.rand_interp(ind_noi)
        anoi: cython.double  = clm.env(ampf) * (1.0 + clm.rand_interp(amp_noi))
        clm.locsig(loc, i, anoi * clm.oscil(carrier, 
                        vib + 
                        inoi * 
                        ((clm.env(indf1) * clm.oscil(fmosc1, (vib * fm1_rat) + fuzz)) +
                        (clm.env(indf2) * clm.oscil(fmosc2, ((vib * fm2_rat) + fuzz))) +
                        (clm.env(indf3) * clm.oscil(fmosc3, ((vib * fm3_rat) + fuzz))))))






#==================================================================================
# The code is ported from Bill Schottstedaet's 'clm-ins.scm' 
# file available at https://ccrma.stanford.edu/software/snd/sndlib/
#==================================================================================

import math
import random
import cython
import numpy as np
import pysndlib.clm as clm
if cython.compiled:
    from cython.cimports.pysndlib import clm

# --------------- gran_synth ---------------- #
@cython.ccall
def gran_synth(startime, duration, audio_freq, grain_dur, grain_interval, amp: cython.double):
    grain_size = math.ceil(max(grain_dur, grain_interval) * clm.get_srate())
    beg: cython.long = clm.seconds2samples(startime)
    end: cython.long = clm.seconds2samples(startime + duration)
    grain_env = clm.make_env([0,0,25,1,75,1,100,0], duration=grain_dur)
    carrier = clm.make_oscil(audio_freq)
    grain = [(clm.env(grain_env) * clm.oscil(carrier)) for i in range(grain_size)]
    grains = clm.make_wave_train(1.0 / grain_interval, grain)

    i: cython.long = 0

    for i in range(beg, end):
        clm.outa(i, amp * clm.wave_train(grains))

if __name__ == '__main__':   
    with clm.Sound( play = True, statistics=True):
        gran_synth(0, 2, 100, .0189, .02, .4)



#==================================================================================
# The code is ported from Bill Schottstedaet's 'generators.scm' 
# file available at https://ccrma.stanford.edu/software/snd/sndlib/
#==================================================================================

import math
import random
import cython
import numpy as np
import pysndlib.clm as clm
import pysndlib.generators as gens
if cython.compiled:
    from cython.cimports.pysndlib import clm
    from cython.cimports.pysndlib import generators as gens

# from generators.scm 
# uses rkoddssb!cos generator

# --------------- glassy ---------------- #
@cython.ccall
def glassy(beg, dur, freq, amp, r=.5):
    r: cython.double = .001**(1.0 / math.floor(clm.get_srate() / (3 * freq)))
    start: cython.long = clm.seconds2samples(beg)
    stop: cython.int  = clm.seconds2samples(beg+dur)
    clang = gens.make_rkoddssb(freq * 2, 1.618 / 2, r)
    clangf = clm.make_env([0,0,.01,1,.1,1,.2,.4,max(.3,dur),0], scaler=amp, duration=dur)
    crf = clm.make_env([0,1,1,0], scaler=r, duration=dur)
    i: cython.long = 0
    for i in range(start, stop):
        clang.r = clm.env(crf)
        clm.outa(i, clm.env(clangf) * gens.rkoddssb(clang, 0.))


if __name__ == '__main__': 
    with clm.Sound(clipped=False, statistics=True, play=True):
        for i in range(10):
            glassy(i*.3, .1, 400. + (100. * i), .5)



# cython: c_string_type=unicode
# cython: c_string_encoding=utf8

#==================================================================================
# The code is part of an attempt at translation of Bill Schottstedaet's sndlib 
# available at https://ccrma.stanford.edu/software/snd/sndlib/
#==================================================================================

#==================================================================================
# In Bill's original he has a nice mechanism for defining generators that I initially
# attempted to replicate but it did not really work well with compiled code 
# so I used a cdef class approach 
#==================================================================================



cimport cython
import numpy as np
cimport numpy as np
#cimport pysndlib.cclm as cclm
cimport pysndlib.clm as clm
from pysndlib.sndlib cimport Sample, Header

np.import_array()

cdef class CLMGenerator:
    cpdef cython.double next(self, cython.double  fm=*)

# --------------- nssb ---------------- #
cdef class Nssb(CLMGenerator):
    cdef public cython.double frequency
    cdef public cython.double ratio
    cdef public cython.int n
    cdef public cython.double angle

    
cpdef cython.double nssb (Nssb gen, cython.double fm=*)
cpdef cython.bint is_nssb(gen)
    
# --------------- nxysin ---------------- #
cdef class Nxysin(CLMGenerator):
    cdef public cython.double frequency
    cdef public cython.double ratio
    cdef public cython.int n
    cdef public cython.double angle
    cdef public cython.double norm

cpdef cython.double nxysin(Nxysin gen, cython.double fm=*)
cpdef cython.bint is_nxysin(gen)

# --------------- nxycos ---------------- #
cdef class Nxycos(CLMGenerator):
    cdef public cython.double frequency 
    cdef public cython.double ratio 
    cdef public cython.int n 
    cdef public cython.double angle 

cpdef cython.double nxycos(Nxycos gen, cython.double fm =*)
cpdef cython.bint is_nxycos(gen)
   
# --------------- nxy1cos ---------------- #
cdef class Nxy1cos(CLMGenerator):
    cdef public cython.double frequency
    cdef public cython.double ratio
    cdef public cython.int n
    cdef public cython.double angle 

cpdef cython.double nxy1cos(Nxy1cos gen, cython.double fm =*)
cpdef cython.bint is_nxy1cos(gen)

# --------------- nxy1sin ---------------- #
cdef class Nxy1sin(CLMGenerator):
    cdef public cython.double frequency
    cdef public cython.double ratio
    cdef public cython.int n
    cdef public cython.double angle

cpdef cython.double nxy1sin(Nxy1sin gen, cython.double fm =*)
cpdef cython.bint is_nxy1sin(gen)

# --------------- Noddcos ---------------- #
cdef class Noddcos(CLMGenerator):
    cdef public cython.double frequency 
    cdef public cython.double ratio 
    cdef public cython.int n 
    cdef public cython.double angle 


cpdef cython.double noddcos(Noddcos gen, cython.double fm=*)
cpdef cython.bint is_noddcos(gen)

# --------------- noddssb ---------------- #
cdef class Noddssb(CLMGenerator):
    cdef public cython.double frequency 
    cdef public cython.double ratio 
    cdef public cython.double n
    cdef public cython.double angle 

cpdef cython.double noddssb(Noddssb gen, cython.double fm=*)
cpdef cython.bint is_noddssb(gen)

# --------------- ncos2 ---------------- #
cdef class Ncos2(CLMGenerator):
    cdef public cython.double frequency 
    cdef public cython.int n
    cdef public cython.double angle 
    
cpdef cython.double ncos2(Ncos2 gen, cython.double fm=*)
cpdef cython.bint is_ncos2(gen)

# --------------- ncos4 ---------------- #
cdef class Ncos4(Ncos2):
    pass

cpdef cython.double ncos4(Ncos4 gen, cython.double fm=*)
cpdef cython.bint is_ncos4(gen)

# --------------- npcos ---------------- #      
cdef class Npcos(CLMGenerator):
    cdef public cython.double frequency 
    cdef public cython.double n
    cdef public cython.double angle 

cpdef cython.double npcos(Npcos gen, cython.double fm=*)
cpdef cython.bint is_npcos(gen)

# --------------- ncos5 ---------------- #  
cdef class Ncos5(CLMGenerator):
    cdef public cython.double frequency 
    cdef public cython.double n
    cdef public cython.double angle 

cpdef cython.double ncos5(Ncos5 gen, cython.double fm=*)
cpdef cython.bint is_ncos5(gen)


# --------------- nsin5 ---------------- #  

cdef class Nsin5(CLMGenerator):
    cdef public cython.double frequency 
    cdef public cython.double n
    cdef public cython.double angle 
    cdef public cython.double norm 

cpdef cython.double nsin5(Nsin5 gen, cython.double fm=*)
cpdef cython.bint is_nsin5(gen)

# --------------- nrsin ---------------- #  
    
cpdef cython.double nrsin(clm.mus_any gen, cython.double fm=*)
cpdef cython.bint is_nrsin(gen)

# 
# --------------- nrcos ---------------- #  
cdef class Nrcos(CLMGenerator):
    cdef public cython.double frequency 
    cdef public cython.double n
    cdef public cython.double r
    cdef public cython.double angle 
    cdef public cython.double rr
    cdef public cython.double r1 
    cdef public cython.double e1 
    cdef public cython.double e2 
    cdef public cython.double norm 
    cdef public cython.bint trouble

cpdef cython.double nrcos(Nrcos gen, cython.double fm=*)
cpdef cython.bint is_nrcos(gen)

# --------------- nrssb ---------------- #
cdef class Nrssb(CLMGenerator):
    cdef public cython.double frequency
    cdef public cython.double ratio 
    cdef public cython.double n
    cdef public cython.double r 
    cdef public cython.double angle
    cdef public cython.double interp
    cdef public cython.double rn
    cdef public cython.double rn1
    cdef public cython.double norm
    cpdef call_interp(self, cython.double fm =*, cython.double interp=*)

cpdef cython.double nrssb(Nrssb gen, cython.double fm=*)
cpdef cython.double nrssb_interp(Nrssb gen, cython.double fm=*,cython.double interp =* )
cpdef cython.bint is_nrssb(gen)

# --------------- nkssb ---------------- #  
cdef class Nkssb(CLMGenerator):
    cdef public cython.double frequency 
    cdef public cython.double ratio
    cdef public cython.double n
    cdef public cython.double angle
    cdef public cython.double norm
    cdef public cython.double interp
    cpdef cython.double call_interp(self, cython.double fm=*, cython.double interp=*)
    
cpdef cython.double nkssb(Nkssb gen, cython.double fm=*)
cpdef cython.double nkssb_interp(Nkssb gen, cython.double fm=*, cython.double interp=*)
cpdef cython.bint is_nkssb(gen)

# --------------- nsincos ---------------- #  
cdef class Nsincos(CLMGenerator):
    cdef public cython.double frequency 
    cdef public cython.double n
    cdef public cython.double angle
    cdef public cython.double n2
    cdef public cython.double cosn
    cdef public cython.double norm
    
cpdef cython.double nsincos(Nsincos gen, cython.double fm=*)
cpdef cython.bint is_nsincos(gen)

# --------------- n1cos ---------------- #  

cdef class N1cos(CLMGenerator):
    cdef public cython.double frequency 
    cdef public cython.double n
    cdef public cython.double angle
    
cpdef cython.double n1cos(N1cos gen, cython.double fm=*)
cpdef cython.bint is_n1cos(gen)

# TODO: --------------- npos1cos ---------------- #  

cdef class Npos1cos(CLMGenerator):
    cdef public cython.double frequency 
    cdef public cython.double n
    cdef public cython.double angle

cpdef cython.double npos1cos(Npos1cos gen, cython.double fm=*)
cpdef cython.bint is_npos1cos(gen)

# --------------- npos3cos ---------------- #  
cdef class Npos3cos(CLMGenerator):
    cdef public cython.double frequency 
    cdef public cython.double n
    cdef public cython.double angle
cpdef cython.double npos3cos(Npos3cos gen, cython.double fm=*)
cpdef cython.bint is_npos3cos(gen)
# --------------- rcos ---------------- #

cdef class Rcos(CLMGenerator):
    cdef public clm.mus_any osc
    cdef public cython.double r
    cdef public cython.double rr
    cdef public cython.double rrp1
    cdef public cython.double rrm1
    cdef public cython.double r2
    cdef public cython.double norm 

cpdef cython.double rcos(Rcos gen, cython.double fm=*)
cpdef cython.bint is_rcos(gen)


# --------------- rssb ---------------- #  

cdef class Rssb(CLMGenerator):
    cdef public cython.double frequency 
    cdef public cython.double ratio
    cdef public cython.double r
    cdef public cython.double angle


    cpdef cython.double call_interp(self, cython.double fm=*, cython.double interp=*)
    
cpdef cython.double rssb(Rssb gen, cython.double fm=*)
cpdef cython.double rssb_interp(Rssb gen, cython.double fm=*, cython.double interp=*)
cpdef cython.bint is_rssb(gen)

# --------------- rxysin ---------------- #  

cdef class Rxysin(CLMGenerator):
    cdef public cython.double frequency 
    cdef public cython.double ratio
    cdef public cython.double r
    cdef public cython.double angle
    cdef public cython.double rr
    cdef public cython.double r2    
 
cpdef cython.double rxysin(Rxysin gen, cython.double fm=*)
cpdef cython.bint is_rxysin(gen)   

# --------------- rxycos ---------------- #  

cdef class Rxycos(CLMGenerator):
    cdef public cython.double frequency 
    cdef public cython.double ratio
    cdef public cython.double r
    cdef public cython.double angle
    cdef public cython.double rr
    cdef public cython.double r2
    cdef public cython.double norm
    
cpdef cython.double rxycos(Rxycos gen, cython.double fm=*)
cpdef cython.bint is_rxycos(gen)   


#  --------------- safe-rxycos ---------------- #  

cdef class SafeRxycos(CLMGenerator):
    cdef public cython.double frequency 
    cdef public cython.double ratio
    cdef public cython.double r
    cdef public cython.double angle
    cdef public cython.double cutoff
    cpdef clamp_rxycos_r(self, cython.double fm=*)
    
cpdef cython.double safe_rxycos(SafeRxycos gen, cython.double fm=*)
cpdef cython.bint is_safe_rxycos(gen)

# --------------- ercos ---------------- #  
cdef class Ercos(CLMGenerator):
    cdef public cython.double frequency 
    cdef public cython.double r
    cdef public cython.double angle
    cdef public cython.double cosh_t
    cdef public cython.double offset
    cdef public cython.double scaler
    cdef public clm.mus_any osc
    
cpdef cython.double ercos(Ercos gen, cython.double fm=*)
cpdef cython.bint is_ercos(gen)

# --------------- erssb ---------------- #  
cdef class Erssb(CLMGenerator):
    cdef public cython.double frequency 
    cdef public cython.double ratio
    cdef public cython.double r
    cdef public cython.double angle

cpdef cython.double erssb(Erssb gen, cython.double fm=*)
cpdef cython.bint is_erssb(gen)


# --------------- r2sin ---------------- #  
# removed


# --------------- r2cos ---------------- #  
cdef class R2cos(CLMGenerator):
    cdef public cython.double frequency 
    cdef public cython.double r
    cdef public cython.double angle

cpdef cython.double r2cos(R2cos gen, cython.double fm=*)
cpdef cython.bint is_r2cos(gen)

# --------------- r2ssb ---------------- #  

cdef class R2ssb(CLMGenerator):
    cdef public cython.double frequency 
    cdef public cython.double ratio
    cdef public cython.double r
    cdef public cython.double angle
    cdef public clm.mus_any osc

cpdef cython.double r2ssb(R2ssb gen, cython.double fm=*)
cpdef cython.bint is_r2ssb(gen)

# --------------- eoddcos ---------------- #  

cdef class Eoddcos(CLMGenerator):
    cdef public cython.double frequency 
    cdef public cython.double r
    cdef public clm.mus_any osc

cpdef cython.double eoddcos(Eoddcos gen, cython.double fm=*)
cpdef cython.bint is_eoddcos(gen)


# --------------- rkcos ---------------- #  

cdef class Rkcos(CLMGenerator):
    cdef public cython.double frequency
    cdef public cython.double r
    cdef public cython.double norm
    cdef public clm.mus_any osc

cpdef cython.double rkcos(Rkcos gen, cython.double fm=*)
cpdef cython.bint is_rkcos(gen)

# --------------- rksin ---------------- #  

cdef class Rksin(CLMGenerator):

    cdef public cython.double frequency
    cdef public cython.double r
    cdef public cython.double angle
    
cpdef cython.double rksin(Rksin gen, cython.double fm=*)
cpdef cython.bint is_rksin(gen)

# --------------- rkssb ---------------- #  

cdef class Rkssb(CLMGenerator):
    cdef public cython.double frequency
    cdef public cython.double r
    cdef public cython.double norm
    cdef public clm.mus_any osc

cpdef cython.double rkcos(Rkcos gen, cython.double fm=*)
cpdef cython.bint is_rkcos(gen)


# --------------- rk!cos ---------------- #  
cdef class Rkfcos(CLMGenerator):

    cdef public cython.double frequency
    cdef public cython.double r
    cdef public cython.double angle
    cdef public cython.double norm
    
cpdef cython.double rkfcos(Rkfcos gen, cython.double fm=*)
cpdef cython.bint is_rkfcos(gen)


# --------------- rk!ssb ---------------- #  

cdef class Rkfssb(CLMGenerator):
    cdef public cython.double frequency 
    cdef public cython.double ratio
    cdef public cython.double r
    cdef public cython.double angle

cpdef cython.double rkfssb(Rkfssb gen, cython.double fm=*)
cpdef cython.bint is_rkfssb(gen)

# TODO: --------------- rxyk!sin ---------------- # 
# remove rxyksin in clm

# TODO: --------------- rxyk!cos ---------------- # 
# remove rxykcos in clm

# --------------- r2k!cos ---------------- # 

cdef class R2kfcos(CLMGenerator):
    cdef public cython.double frequency 
    cdef public cython.double r
    cdef public cython.double k 
    cdef public cython.double rr1
    cdef public cython.double r2
    cdef public clm.mus_any osc
    cdef public cython.double norm
    
cpdef cython.double r2kfcos(R2kfcos gen, cython.double fm=*)
cpdef cython.bint is_r2kfcos(gen)

# --------------- k2sin ---------------- # 

cdef class K2sin(CLMGenerator):
    cdef public cython.double frequency 
    cdef public cython.double angle
    
cpdef cython.double k2sin(K2sin gen, cython.double fm=*)
cpdef cython.bint is_k2sin(gen)

# --------------- k2cos ---------------- # 

cdef class K2cos(CLMGenerator):
    cdef public cython.double frequency 
    cdef public cython.double angle
    
cpdef cython.double k2cos(K2cos gen, cython.double fm=*)
cpdef cython.bint is_k2cos(gen)

# --------------- k2ssb ---------------- # 

cdef class K2ssb(CLMGenerator):
    cdef public cython.double frequency 
    cdef public cython.double ratio
    cdef public cython.double angle
    
cpdef cython.double k2ssb(K2ssb gen, cython.double fm=*)
cpdef cython.bint is_k2ssb(gen)

# --------------- dblsum ---------------- # 

cdef class Dblsum(CLMGenerator):
    cdef public cython.double frequency 
    cdef public cython.double r
    cdef public cython.double angle
    
cpdef cython.double dblsum(Dblsum gen, cython.double fm=*)
cpdef cython.bint is_dblsum(gen)


# --------------- rkoddssb ---------------- #
cdef class Rkoddssb(CLMGenerator):
    cdef public cython.double frequency 
    cdef public cython.double ratio 
    cdef public cython.double r
    cdef public cython.double rr1
    cdef public cython.double angle 
    cdef public cython.double norm 
    
cpdef cython.double rkoddssb(Rkoddssb gen, cython.double fm=*)
cpdef cython.bint is_rkoddssb(gen)

# --------------- krksin ---------------- # 

cdef class Krksin(CLMGenerator):
    cdef public cython.double frequency 
    cdef public cython.double r
    cdef public cython.double angle
    
cpdef cython.double krksin(Krksin gen, cython.double fm=*)
cpdef cython.bint is_krksin(gen)

# --------------- abssin ---------------- # 

cdef class Absin(CLMGenerator):
    cdef public cython.double frequency 
    cdef public clm.mus_any osc
    
cpdef cython.double absin(Absin gen, cython.double fm=*)
cpdef cython.bint is_absin(gen)

# --------------- abscos ---------------- # 

cdef class Abcos(CLMGenerator):
    cdef public cython.double frequency 
    cdef public cython.double a
    cdef public cython.double b
    cdef public cython.double angle
    cdef public cython.double ab
    cdef public cython.double norm
    
cpdef cython.double abcos(Abcos gen, cython.double fm=*)
cpdef cython.bint is_abcos(gen)

# --------------- r2k2cos ---------------- # 

cdef class R2k2cos(CLMGenerator):
    cdef public cython.double frequency 
    cdef public cython.double r
    cdef public cython.double angle
    cdef cython.double r2k2cos_norm(self, cython.double a)
    
cpdef cython.double r2k2cos(R2k2cos gen, cython.double fm=*)
cpdef cython.bint is_r2k2cos(gen)

# --------------- blsaw ---------------- # 

cdef class Blsaw(CLMGenerator):
    cdef public cython.double frequency 
    cdef public cython.double n
    cdef public cython.double r
    cdef public cython.double angle
    
cpdef cython.double blsaw(Blsaw gen, cython.double fm=*)
cpdef cython.bint is_blsaw(gen)


# TODO: add I variant--------------- asyfm ---------------- # 

cdef class Asyfm(CLMGenerator):
    cdef public cython.double frequency 
    cdef public cython.double ratio
    cdef public cython.double r
    cdef public cython.double index
    cdef public cython.double angle
    cpdef cython.double next_i(self, cython.double fm =*)
    cpdef cython.double next_j(self, cython.double fm =*)
    
cpdef cython.double asyfm(Asyfm gen, cython.double fm=*)
cpdef cython.double asyfm_i(Asyfm gen, cython.double fm=*)
cpdef cython.double asyfm_j(Asyfm gen, cython.double fm=*)
cpdef cython.bint is_asyfm(gen)
# 
# --------------- bess ---------------- # 

cdef class Bess(CLMGenerator):
    cdef public cython.double frequency 
    cdef public cython.int n
    cdef public cython.double angle
    cdef public cython.double norm
    
cpdef cython.double bess(Bess gen, cython.double fm=*)
cpdef cython.bint is_bess(gen)


# --------------- jjcos ---------------- # 
cdef class Jjcos(CLMGenerator):
    cdef public cython.double frequency 
    cdef public cython.double r
    cdef public cython.double a
    cdef public cython.double k
    cdef public cython.double angle
    
cpdef cython.double jjcos(Jjcos gen, cython.double fm=*)
cpdef cython.bint is_jjcos(gen)

# --------------- j0evencos ---------------- # 
cdef class J0evencos(CLMGenerator):
    cdef public cython.double frequency 
    cdef public cython.double index
    cdef public cython.double angle
    
cpdef cython.double j0evencos(J0evencos gen, cython.double fm=*)
cpdef cython.bint is_j0evencos(gen)

# --------------- j2cos ---------------- # 


cdef class J2cos(CLMGenerator):
    cdef public cython.double frequency 
    cdef public cython.int n
    cdef public cython.double r
    cdef public cython.double angle
    cdef public cython.double norm
    
cpdef cython.double j2cos(J2cos gen, cython.double fm=*)
cpdef cython.bint is_j2cos(gen)


# --------------- jpcos ---------------- # 

cdef class Jpcos(CLMGenerator):
    cdef public cython.double frequency 
    cdef public cython.double r
    cdef public cython.double a
    cdef public cython.double k
    cdef public cython.double angle
    
cpdef cython.double jpcos(Jpcos gen, cython.double fm=*)
cpdef cython.bint is_jpcos(gen)
# 


# --------------- jncos ---------------- # 
cdef class Jncos(CLMGenerator):
    cdef public cython.double frequency 
    cdef public cython.double r
    cdef public cython.double a
    cdef public cython.int n
    cdef public cython.double angle
    cdef public cython.double ra
    
cpdef cython.double jncos(Jncos gen, cython.double fm=*)
cpdef cython.bint is_jncos(gen)

# --------------- j0j1cos  ---------------- # 

cdef class J0j1cos(CLMGenerator):
    cdef public cython.double frequency 
    cdef public cython.double index
    cdef public cython.double angle
    
cpdef cython.double j0j1cos(J0j1cos gen, cython.double fm=*)
cpdef cython.bint is_j0j1cos(gen)

# --------------- jycos  ---------------- # 

cdef class Jycos(CLMGenerator):
    cdef public cython.double frequency 
    cdef public cython.double r
    cdef public cython.double a
    cdef cython.double angle
    
cpdef cython.double jycos(Jycos gen, cython.double fm=*)
cpdef cython.bint is_jycos(gen)

# --------------- jcos  ---------------- # 
cdef class Jcos(CLMGenerator):
    cdef public cython.double frequency 
    cdef public cython.int n
    cdef public cython.double r
    cdef public cython.double a
    
cpdef cython.double jcos(Jcos gen, cython.double fm=*)
cpdef cython.bint is_jcos(gen)
 
# --------------- blackman  ---------------- # 
cdef class Blackman(CLMGenerator):
    cdef public cython.double frequency 
    cdef public cython.int n
    cdef public clm.mus_any osc
    cpdef mus_reset(self)
    
cpdef cython.double blackman(Blackman gen, cython.double fm=*)
cpdef cython.bint is_blackman(gen)

# --------------- fmssb ---------------- #

cdef class Fmssb(CLMGenerator):
    cdef public cython.double frequency 
    cdef public cython.double ratio
    cdef public cython.double idx
    cdef public cython.double angle 

cpdef cython.double fmssb(Fmssb gen, cython.double fm=*)
cpdef cython.bint is_fmssb(gen)

# --------------- k3sin  ---------------- # 

cdef class K3sin(CLMGenerator):
    cdef public cython.double frequency 
    cdef public np.ndarray coeffs
    cdef public cython.double angle 
    
cpdef cython.double k3sin(K3sin gen, cython.double fm=*)
cpdef cython.bint is_k3sin(gen)


# --------------- izcos  ---------------- # # 
cdef class Izcos(CLMGenerator):
    cdef public cython.double frequency 
    cdef public cython.double r
    cdef public cython.double angle 
    cdef public cython.double dc
    cdef public cython.double norm
    cdef public cython.double inorm

cpdef cython.double izcos(Izcos gen, cython.double fm=*)
cpdef cython.bint is_izcos(gen)
# 
# --------------- adjustable-square-wave ---------------- # 
# Note: modulating the duty factor will not work
cdef class AdjustableSquareWave(CLMGenerator):
    cdef public cython.double frequency 
    cdef public cython.double duty_factor
    cdef public cython.double amplitude
    cdef public cython.double sum
    cdef public clm.mus_any p1
    cdef public clm.mus_any p2
    
cpdef cython.double adjustable_square_wave(AdjustableSquareWave gen, cython.double fm=*)
cpdef cython.bint is_adjustable_square_wave(gen)
# # 
# --------------- adjustable-triangle-wave ---------------- # 
# 
cdef class AdjustableTriangleWave(CLMGenerator):
    cdef public cython.double frequency 
    cdef public cython.double duty_factor
    cdef public cython.double amplitude
    cdef public clm.mus_any gen
    cdef public cython.double top
    cdef public cython.double mtop
    cdef public cython.double scl
    cdef public cython.double val
    
cpdef cython.double adjustable_triangle_wave(AdjustableTriangleWave gen, cython.double fm=*)
cpdef cython.bint is_adjustable_triangle_wave(gen)
# 
# # 
# --------------- adjustable-sawtooth-wave ---------------- # 

cdef class AdjustableSawtoothWave(CLMGenerator):
    cdef public cython.double frequency 
    cdef public cython.double duty_factor
    cdef public cython.double amplitude
    cdef public clm.mus_any gen
    cdef public cython.double top
    cdef public cython.double mtop
    cdef public cython.double scl
    cdef public cython.double val
    
cpdef cython.double adjustable_sawtooth_wave(AdjustableSawtoothWave gen, cython.double fm=*)
cpdef cython.bint is_adjustable_sawtooth_wave(gen)
# # 
# TODO: --------------- adjustable-oscil-wave ---------------- # 

cdef class AdjustableOscil(CLMGenerator):
    cdef public cython.double frequency 
    cdef public cython.double duty_factor
    cdef public cython.double amplitude
    cdef public clm.mus_any gen
    cdef public cython.double top
    cdef public cython.double mtop
    cdef public cython.double scl
    cdef public cython.double val
    
cpdef cython.double adjustable_oscil(AdjustableOscil gen, cython.double fm=*)
cpdef cython.bint is_adjustable_oscil(gen)


# --------------- make-table-lookup-with-env ---------------- # 
cpdef make_table_lookup_with_env(frequency, pulse_env, size=*)

# --------------- make-wave-train-with-env ---------------- # 
cpdef make_wave_train_with_env(frequency, pulse_env, size=*)

# --------------- round-interp ---------------- # 
cdef class RoundInterp(CLMGenerator):
    cdef public cython.double frequency 
    cdef public cython.int n
    cdef public cython.double amplitude
    cdef public clm.mus_any rnd
    cdef public clm.mus_any flt
    
cpdef cython.double round_interp(RoundInterp gen, cython.double fm=*)
cpdef cython.bint is_round_interp(gen)


# --------------- env-any functions ---------------- # 

# --------------- sine-env ---------------- # 
cpdef cython.double sine_env(e)

# --------------- square-env ---------------- # 
cpdef cython.double square_env(e)

# --------------- blackman4-env ---------------- # 
cpdef cython.double blackman4_env(e)

# --------------- multi-expt-env ---------------- # 
# going to be slow
#def multi_expt_env(e, expts):

# --------------- run-with-fm-and-pm ---------------- # 
# skip for now

# TODO: This is not working correctly --------------- nchoosekcos ---------------- # 
cdef class Nchoosekcos(CLMGenerator):
    cdef public cython.double frequency 
    cdef public cython.double ratio
    cdef public cython.int n
    cdef public cython.double angle
    
cpdef cython.double nchoosekcos(Nchoosekcos gen, cython.double fm=*)
cpdef cython.bint is_nchoosekcos(gen)
# 
# 
# # 
# TODO: --------------- sinc-train ---------------- # 

cdef class SincTrain(CLMGenerator):
    cdef public cython.double frequency 
    cdef public cython.int n
    cdef public cython.double angle
    cdef public cython.double original_n
    cdef public cython.double original_frequency
    
cpdef cython.double sinc_train(SincTrain gen, cython.double fm=*)
cpdef cython.bint is_sinc_train(gen)

# --------------- pink-noise ---------------- # 
cdef class PinkNoise(CLMGenerator):
    cdef public cython.int n
    cdef public np.ndarray data 

cpdef cython.double pink_noise(PinkNoise gen, cython.double fm=*)
cpdef cython.bint is_pink_noise(gen)

# --------------- brown-noise ---------------- # 
cdef class BrownNoise(CLMGenerator):
    cdef public cython.double frequency
    cdef public cython.double amplitude
    cdef public cython.double prev
    cdef public cython.double sum
    cdef public clm.mus_any gr
    
cpdef cython.double brown_noise(BrownNoise gen, cython.double fm=*)
cpdef cython.bint is_brown_noise(gen)
# 
# 
# --------------- green-noise ---------------- # 
cdef class GreenNoise(CLMGenerator):
    cdef public cython.double frequency
    cdef public cython.double amplitude
    cdef public cython.double prev
    cdef public cython.double sum
    cdef public clm.mus_any gr
    cdef public cython.double low
    cdef public cython.double high
#     
# TODO: This is not working properly--------------- green-noise-interp ---------------- # 
cdef class GreenNoiseInterp(CLMGenerator):
    cdef public cython.double frequency
    cdef public cython.double amplitude
    cdef public cython.double sum
    cdef public cython.double low
    cdef public cython.double high
    cdef public cython.double dv
    cdef public cython.double incr
    cdef public cython.double angle
# 
# 
# 
# --------------- moving-sum ---------------- # 

cdef class MovingSum(CLMGenerator):
    cdef public cython.int n
    cdef public clm.mus_any gen

cpdef cython.double moving_sum(MovingSum gen, cython.double insig=*)
cpdef cython.bint is_moving_sum(gen)

# --------------- moving-variance ---------------- # 

cdef class MovingVariance(CLMGenerator):
    cdef public cython.int n
    cdef public clm.mus_any gen1
    cdef public clm.mus_any gen2

cpdef cython.double moving_variance(MovingVariance gen, cython.double insig=*)
cpdef cython.bint is_moving_variance(gen)

# --------------- moving-rms ---------------- # 

cdef class MovingRMS(CLMGenerator):
    cdef public cython.int n
    cdef public clm.mus_any gen

cpdef cython.double moving_rms(MovingRMS gen, cython.double insig=*)
cpdef cython.bint is_moving_rms(gen)

# --------------- moving-length ---------------- # 

cdef class MovingLength(MovingRMS):
    pass
    
cpdef cython.double moving_length(MovingLength gen, cython.double insig=*)
cpdef cython.bint is_moving_length(gen)

# --------------- weighted-moving-average---------------- # 

cdef class WeightedMovingAverage(CLMGenerator):
    cdef public cython.int n
    cdef public clm.mus_any dly
    cdef public cython.double num
    cdef public cython.double sum
    cdef public cython.double y
    cdef public cython.double den

cpdef cython.double weighted_moving_average(WeightedMovingAverage gen, cython.double insig=*)
cpdef cython.bint is_weighted_moving_average(gen)

# --------------- exponentially-weighted-moving-average---------------- # 

cdef class ExponentiallyWeightedMovingAverage(CLMGenerator):
    cdef public cython.int n
    cdef public clm.mus_any gen

cpdef cython.double exponentially_weighted_moving_average(ExponentiallyWeightedMovingAverage gen, cython.double insig=*)
cpdef cython.bint is_exponentially_weighted_moving_average(gen)

# --------------- polyoid ---------------- # 

cpdef cython.double polyoid(clm.mus_any gen, cython.double fm=*)
cpdef cython.bint is_polyoid(gen)


# # # TODO: these require potentially loading a file so
# # # wondering good way to do this. using maybe binary numpy files

# TODO: --------------- noid ---------------- # 
# TODO: --------------- knoid ---------------- # 
# TODO: --------------- roid ---------------- # 


# Removed --------------- waveshape ---------------- # 


# --------------- tanhsin ---------------- # 
cdef class Tanhsin(CLMGenerator):
    cdef public cython.double frequency 
    cdef public cython.double r
    cdef public clm.mus_any osc 

cpdef cython.double tanhsin(Tanhsin gen, cython.double fm=*)
cpdef cython.bint is_tanhsin(gen)   
# 
# --------------- moving-fft ---------------- # 

# cdef class MovingFFT(CLMGenerator):
#     cdef public clm.mus_any reader
#     cdef public cython.int n 
#     cdef public cython.int hop
#     cdef public cython.int outctr
#     cdef public np.ndarray rl
#     cdef public np.ndarray im
#     cdef public np.ndarray data
#     cdef public np.ndarray window
#     cdef last_moving_fft_window
#     cpdef mus_run(self, arg1, arg1)
# 
# cpdef cython.double moving_fft(MovingFFT gen)
# cpdef cython.bint is_moving_fft(gen)
# 
# # --------------- moving-spectrum ---------------- # 
# 
# 
# cdef class MovingSpectrum(CLMGenerator):
#     cdef public cython.int n 
#     cdef public cython.int hop
#     cdef public cython.int outctr
#     cdef public np.ndarray amps
#     cdef public np.ndarray phases
#     cdef public np.ndarray amp_incs
#     cdef public np.ndarray freqs
#     cdef public np.ndarray freq_incs
#     cdef public np.ndarray new_freq_incs
#     cdef public np.ndarray window
#     cdef public np.ndarray data
#     cdef public cython.long dataloc
# 
# cpdef cython.double moving_spectrum(MovingSpectrum gen)
# cpdef cython.bint is_moving_spectrum(gen)
# 
# 
# --------------- moving-scentroid ---------------- # 
# cdef class MovingScentroid(CLMGenerator):
#     cdef public cython.double dbfloor
#     cdef public cython.double rfreq
#     cdef public cython.int size
#     cdef public cython.int hop
#     cdef public cython.int outctr
#     cdef public cython.double curval
#     cdef public cython.double binwidth
#     cdef public np.ndarray rl
#     cdef public np.ndarray im
#     cdef public clm.mus_any dly
#     cdef public clm.mus_any rms
#     cdef public cython.double x
#     
# cpdef cython.double moving_scentroid(MovingScentroid gen, cython.double x=*)
# cpdef cython.bint is_moving_scentroid(gen)

# # --------------- moving-autocorrelation ---------------- # 
# 
# cdef class MovingAutocorrelation(CLMGenerator):
#     cdef public cython.int n 
#     cdef public cython.int hop
#     cdef public cython.int outctr
#     cdef public np.ndarray rl
#     cdef public np.ndarray im
#     cdef public np.ndarray data
# 
# cpdef cython.double moving_autocorrelation(MovingAutocorrelation gen)
# cpdef cython.bint is_moving_autocorrelationt(gen)
# 
# # --------------- moving-pitch ---------------- # 
# 
# cdef class MovingPitch(CLMGenerator):
#     cdef public cython.int n 
#     cdef public cython.int hop
#     cdef public clm.mus_any ac
#     cdef public cython.double val
# 
# cpdef cython.double moving_pitch(MovingPitch gen)
# cpdef cython.bint is_moving_pitch(gen)
# 
# # --------------- flocsig ---------------- # 
# 
# cdef class Flocsig(CLMGenerator):
#     cdef public cython.double reverb_amount
#     cdef public cython.double frequency
#     cdef public cython.double amplitude
#     cdef public cython.double offset
#     cdef public cython.int maxd
#     cdef public np.ndarray out1
#     cdef public np.ndarray out2
#     cdef public cython.long outloc
#     cdef public clm.mus_any ri
#     cdef public cython.long samp
#     cdef public cython.double insig
#     
# 
# cpdef cython.double flocsig(Flocsig gen, cython.long samp, cython.double insig)
# cpdef cython.bint is_flocsig(gen)
# """



#==================================================================================
# The code is ported from Bill Schottstedaet's 'generators.scm' 
# file  available at https://ccrma.stanford.edu/software/snd/sndlib/
#==================================================================================

import math
import random
import cython
import numpy as np
import pysndlib.clm as clm
import pysndlib.generators as gens
if cython.compiled:
    from cython.cimports.pysndlib import clm
    from cython.cimports.pysndlib import generators as gens

# from generators.scm 
# uses r2k!co generator

# --------------- pianoy ---------------- #
@cython.ccall
def pianoy(start, dur, freq, amp):
    gen = gens.make_r2kfcos(freq, r=.5, k=3)
    ampf = clm.make_env([0,0,.01,1,.03,1,1,.15, max(2,dur), 0.0], base=32, scaler=amp, duration=dur)
    beg: cython.long = clm.seconds2samples(start)
    end: cython.long = clm.seconds2samples(start+dur)
    i:  cython.long = 0
    for i in range(beg, end):
        clm.outa(i, clm.env(ampf) *gens.r2kfcos(gen))


if __name__ == '__main__': 
    with clm.Sound(statistics=True, play=True):
        pianoy(0,3,400,.5)



#==================================================================================
# The code is ported from Bill Schottstedaet's 'clm-ins.scm' 
# file  available at https://ccrma.stanford.edu/software/snd/sndlib/
#==================================================================================

import math
import random
import cython
import numpy as np
import pysndlib.clm as clm
import pysndlib.env as env
if cython.compiled:
    from cython.cimports.pysndlib import clm

# --------------- fm_trumpet ---------------- #

# ;;; Dexter Morrill's FM-trumpet:
# ;;; from CMJ feb 77 p51

@cython.ccall
def fm_trumpet(startime, dur, frq1=250, frq2=1500, amp1=.5, amp2=.1, 
    ampatt1=.03, ampdec1=.35,
    ampatt2=.03, ampdec2=.3,
    modfrq1=250,
    modind11=0.0,
    modind12=2.66,
    modfrq2=250.,
    modind21=0.0,
    modind22=1.8,
    rvibamp=.007,
    rvibfrq=125.0,
    vibamp=.007,
    vibfrq=7.0,
    vibatt=.6,
    vibdec=.2,
    frqskw=.03,
    frqatt=.06,
    ampenv1=[0,0,25,1,75,.9,100,0],
    ampenv2=[0,0,25,1,75,.9,100,0],
    indenv1=[0,0,25,1,75,.9,100,0],
    indenv2=[0,0,25,1,75,.9,100,0],
    degree=0.0,
    distance=1.0,
    reverb_amount=.005):
    
    dec_01 = max(75, (100 * (1.0 - (.01 / dur))))
    beg: cython.long = clm.seconds2samples(startime)
    end: cython.long = clm.seconds2samples(startime + dur)
    loc = clm.make_locsig(degree, distance, reverb_amount)
    vibe = env.stretch_envelope([0,1,25,.1, 75, 0, 100, 0], 25, min((100 * (vibatt / dur)), 45), 75, max((100 * (1.0 - (vibdec / dur)), 55)))

    per_vib_f = clm.make_env(vibe, scaler=vibamp, duration=dur)
    ran_vib = clm.make_rand_interp(rvibfrq, rvibamp)
    per_vib = clm.make_oscil(vibfrq)
    frqe = env.stretch_envelope([0,0,25,1,75,1,100,0], 25, min(25, (100 * (frqatt / dur))), 75, dec_01)
    frq_f = clm.make_env(frqe, scaler=frqskw, offset=1.0, duration=dur)
    ampattpt1 = min(25, (100 * (ampatt1 / dur)))
    ampdecpt1 = max(75, (100 * (1.0 - (ampdec1 / dur))))
    ampattpt2 = min(25, (100 * (ampatt2 / dur)))
    ampdecpt2 = max(75, (100 * (1.0 - (ampdec2 / dur))))
    
    mod1_f = clm.make_env(env.stretch_envelope(indenv1, 25, ampattpt1, 75, dec_01), scaler=(modfrq1 * (modind12 - modind11)), duration=dur)
    mod1 = clm.make_oscil(0.0)
    car1 = clm.make_oscil(0.0)
    car1_f = clm.make_env(env.stretch_envelope(ampenv1, 25, ampattpt1, 75, ampdecpt1), scaler=amp1, duration=dur)
    
    mod2_f = clm.make_env(env.stretch_envelope(indenv2, 25, ampattpt2, 75, dec_01), scaler=(modfrq2 * (modind22 - modind21)), duration=dur)

    mod2 = clm.make_oscil(0.0)
    car2 = clm.make_oscil(0.0)
    car2_f = clm.make_env(env.stretch_envelope(ampenv2, 25, ampattpt2, 75, ampdecpt2), scaler=amp2, duration=dur)
    
    i: cython.long = 0
    frq_change: cython.double = 0.0
    
    for i in range(beg, end):
        frq_change = clm.hz2radians((1.0 + clm.rand_interp(ran_vib)) * 
                            (1.0 + (clm.env(per_vib_f) * clm.oscil(per_vib))) *
                            clm.env(frq_f))
                            
                            
        clm.locsig(loc, i, clm.env(car1_f) * clm.oscil(car1, frq_change * (frq1 + clm.env(mod1_f) * clm.oscil(mod1, modfrq1 * frq_change))) +
                        clm.env(car2_f) * clm.oscil(car2, frq_change * (frq2 + clm.env(mod2_f) * clm.oscil(mod2, modfrq2 * frq_change))))
if __name__ == '__main__': 
    with clm.Sound(play=True, statistics=True):
        fm_trumpet(0, 3, 250, 1500, .5, .1, degree=45)
        fm_trumpet(2.9, 3, 350, 1500, .5, .1, degree=45)    




#==================================================================================
# The code is ported from Bill Schottstedaet's 'clm-ins.scm' 
# file  available at https://ccrma.stanford.edu/software/snd/sndlib/
#==================================================================================

import math
import random
import cython
import numpy as np
import pysndlib.clm as clm
if cython.compiled:
    from cython.cimports.pysndlib import clm
# --------------- hammondoid ---------------- #
#   ;; from Perry Cook's BeeThree.cpp
@cython.ccall
def hammondoid(beg, dur, freq, amp):
    osc0  = clm.make_oscil(freq * .999)
    osc1  = clm.make_oscil(freq * 1.997)
    osc2  = clm.make_oscil(freq * 3.006)
    osc3  = clm.make_oscil(freq * 6.009)
    ampenv1  = clm.make_env([0,0,.005, 1, dur-.008, 1, dur, 0], duration=dur)
    ampenv2  = clm.make_env([0,0,.005, 1, dur, 0], duration=dur, scaler=(.5 * .75 * amp))
    g0: cython.double = .25 * .75 * amp
    g1: cython.double = .25 * .75 * amp
    g2: cython.double = .5 * amp
    st: cython.long = clm.seconds2samples(beg)
    nd: cython.long = clm.seconds2samples(beg + dur)
    
    i: cython.long = 0
        
    for i in range(st, nd):
        clm.outa(i, (clm.env(ampenv1) * ((g0 * clm.oscil(osc0)) + (g1 * clm.oscil(osc1)) + (g2 * clm.oscil(osc2)))) +
                    (clm.env(ampenv2)*clm.oscil(osc3)))



from setuptools import Extension, setup
from Cython.Build import cythonize
import Cython.Compiler.Options
Cython.Compiler.Options.annotate = False
import numpy
import sys
import os
import glob

# 
# linking to static lib
# 
if sys.platform.startswith("darwin"):
    os.environ['LDFLAGS'] = '-framework CoreAudio -framework CoreFoundation -framework CoreMIDI'
    os.environ['CFLAGS'] = '-arch x86_64 -arch arm64'
    
    extensions = [
        Extension("pysndlib.sndlib", ["src/pysndlib/sndlib.pyx"], 
            extra_objects=["./sndlib/libsndlib.a"], 
            include_dirs=[numpy.get_include(), './sndlib'],
            extra_compile_args=['-Wno-parentheses-equality', '-Wno-unreachable-code-fallthrough' ],
            define_macros=[("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")]),
        Extension("pysndlib.clm", ["src/pysndlib/clm.pyx"], 
            extra_objects=["./sndlib/libsndlib.a"], 
            include_dirs=[numpy.get_include(), './sndlib'], 
            libraries=["m"],
            extra_compile_args=['-Wno-parentheses-equality', '-Wno-unreachable-code-fallthrough'],
            define_macros=[("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")]),
        Extension("pysndlib.env", ["src/pysndlib/env.pyx"], 
            extra_objects=["./sndlib/libsndlib.a"], 
            include_dirs=[numpy.get_include(), './sndlib'], 
            extra_compile_args=['-Wno-parentheses-equality', '-Wno-unreachable-code-fallthrough'],
            define_macros=[("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")]),
        Extension("pysndlib.dsp", ["src/pysndlib/dsp.pyx"], 
            extra_objects=["./sndlib/libsndlib.a"], 
            include_dirs=[numpy.get_include(), './sndlib'], 
            extra_compile_args=['-Wno-parentheses-equality', '-Wno-unreachable-code-fallthrough'],
            define_macros=[("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")]),
        Extension("pysndlib.generators", ["src/pysndlib/generators.pyx"], 
            extra_objects=["./sndlib/libsndlib.a"], 
            include_dirs=[numpy.get_include(), './sndlib'], 
            extra_compile_args=['-Wno-parentheses-equality', '-Wno-unreachable-code-fallthrough'],
            define_macros=[("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")]),
        Extension("*", ["src/pysndlib/instruments/*.py"], 
            extra_objects=["./sndlib/libsndlib.a"], 
            include_dirs=[numpy.get_include(), './sndlib'], 
            extra_compile_args=['-Wno-parentheses-equality', '-Wno-unreachable-code-fallthrough'],
            define_macros=[("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")]),
    ]


if sys.platform.startswith("linux"):
    os.environ['LDFLAGS'] = '-lm -ldl'

    extensions = [
        Extension("pysndlib.sndlib", ["src/pysndlib/sndlib.pyx"], 
            extra_objects=["./sndlib/libsndlib.a"], 
            include_dirs=[numpy.get_include(), './sndlib'],
            extra_compile_args=['-Wno-parentheses-equality', '-Wno-unreachable-code-fallthrough' ],
            define_macros=[("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")]),
        Extension("pysndlib.clm", ["src/pysndlib/clm.pyx"], 
            extra_objects=["./sndlib/libsndlib.a"], 
            include_dirs=[numpy.get_include(), './sndlib'], 
            libraries=["m"],
            extra_compile_args=['-Wno-parentheses-equality', '-Wno-unreachable-code-fallthrough'],
            define_macros=[("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")]),
        Extension("pysndlib.env", ["src/pysndlib/env.pyx"], 
            extra_objects=["./sndlib/libsndlib.a"], 
            include_dirs=[numpy.get_include(), './sndlib'], 
            extra_compile_args=['-Wno-parentheses-equality', '-Wno-unreachable-code-fallthrough'],
            define_macros=[("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")]),
        Extension("pysndlib.dsp", ["src/pysndlib/dsp.pyx"], 
            extra_objects=["./sndlib/libsndlib.a"], 
            include_dirs=[numpy.get_include(), './sndlib'], 
            extra_compile_args=['-Wno-parentheses-equality', '-Wno-unreachable-code-fallthrough'],
            define_macros=[("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")]),
        Extension("pysndlib.generators", ["src/pysndlib/generators.pyx"], 
            extra_objects=["./sndlib/libsndlib.a"], 
            include_dirs=[numpy.get_include(), './sndlib'], 
            extra_compile_args=['-Wno-parentheses-equality', '-Wno-unreachable-code-fallthrough'],
            define_macros=[("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")]),
        Extension("*", ["src/pysndlib/instruments/*.py"], 
            extra_objects=["./sndlib/libsndlib.a"], 
            include_dirs=[numpy.get_include(), './sndlib'], 
            extra_compile_args=['-Wno-parentheses-equality', '-Wno-unreachable-code-fallthrough'],
            define_macros=[("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")]),
    ]


if __name__ == '__main__':
    
    setup(
        zip_safe=False,
        name = 'pysndlib',
        package_data={'pysndlib': ["*.pyx", "*.py", "*.snd", "*.aiff", "*.wav", "*.pxd"]},
        ext_modules = cythonize(extensions, compiler_directives={'language_level': '3', 'embedsignature' : False})
    )


